# coding: utf-8

# Standard Python libraries
from typing import Any

def dict_insert(d: dict,
                key: Any,
                value: Any,
                **kwargs):
    """
    Adds a new key, value pair into a dictionary by inserting it
    relative to a known key.
    
    Parameters
    ----------
    d : dict
        The dict to add the new key, value to.
    key : any
        The new dict key to insert.
    value : any
        The dict value to assign to d[key].
    before : any, optional
        If given, then the new key will be inserted before this key.  Cannot be
        given with after.
    after : any, optional
        If given, then the new key will be inserted after this key.  Cannot be
        given with before
    """
    if len(kwargs) == 0:
        # Set normally if no kwargs given.
        d[key] = value
        return

    elif len(kwargs) == 1:
        if 'before' in kwargs:
            try:
                i = list(d.keys()).index(kwargs['before'])
            except:
                # Set normally if ref key not found
                d[key] = value
                return
        elif 'after' in kwargs:
            try:
                i = list(d.keys()).index(kwargs['after']) + 1
            except:
                # Set normally if ref key not found
                d[key] = value
                return
        else:
            raise TypeError(f"dict_insert() got an unexpected keyword argument '{list(kwargs.keys())[0]}'")
    else:
        raise ValueError('Only one kwarg (before or after) can be given')

    # Remove d if currently in d
    if key in d:
        del d[key]

    # Pop all keys after previous
    newdict = {}
    for k in list(d.keys())[i:]:
        newdict[k] = d.pop(k)

    # Set key, value
    d[key] = value

    # Add popped keys back in
    d.update(newdict)



from IPython.display import display, Markdown

from .record import load_record


def querydoc(style: str,
             render: bool = False):
    """
    Generate the querydoc content for a record style.

    Parameters
    ----------
    style : str
        The record style
    render : bool
        If True then the query doc will be rendered for IPython environments.
        If False (default) then it will be returned as a str.
    """

    # Load an empty record
    record = load_record(style)

    # Return/render the querydoc
    if render:
        display(Markdown(record.querydoc))
    else:
        return record.querydoc


@property
def ms(self) -> float:
    """millisecond"""
    return 1e-3 * self.s

@property
def us(self) -> float:
    """microsecond"""
    return 1e-6 * self.s

@property
def ns(self) -> float:
    """nanosecond"""
    return 1e-9 * self.s

@property
def ps(self) -> float:
    """picosecond"""
    return 1e-12 * self.s

@property
def fs(self) -> float:
    """femtosecond"""
    return 1e-15 * self.s

@property
def minute(self) -> float:
    """minute"""
    return 60. * self.s

@property
def hour(self) -> float:
    """hour"""
    return 60. * self.minute

@property
def day(self) -> float:
    """solar day"""
    return 24. * self.hour

@property
def week(self) -> float:
    """week"""
    return 7. * self.day

@property
def year(self) -> float:
    """sidereal year"""
    return 365.256363004 * self.day


import builtins
from unittest.mock import patch
import shutil
from pathlib import Path

from pytest import raises

import yabadaba


class TestSettings():

    @property
    def directoryname(self):
        """str : The directoryname value used for initialization"""
        return 'yabadaba_testing'
    
    @property
    def fname(self):
        """str : The filename value used for initialization"""
        return 'yabadaba_testing_settings.json'

    @property
    def defaultdirectory(self):
        """Path : What the defaultdirectory value should be"""
        return Path(Path.home(), self.directoryname)

    @property
    def directory(self):
        """Path : What the directory value should be"""
        return self.__directory

    @directory.setter
    def directory(self, value):
        self.__directory = Path(value).resolve()

    @property
    def defaultfilename(self):
        """Path : What the defaultfilename value should be"""
        return Path(self.defaultdirectory, self.fname)

    @property
    def filename(self):
        """Path : What the filename value should be"""
        return Path(self.directory, self.fname)

    @property
    def settings(self):
        """yabadaba.Settings.Settings : A settings object to use for testing"""
        if self.defaultdirectory.exists():
            shutil.rmtree(self.defaultdirectory)
        return yabadaba.Settings.Settings(directoryname=self.directoryname,
                                          filename=self.fname)
        
    def test_initial(self):
        """Initial testing"""

        # Initialize a settings object
        settings = self.settings

        # Check status of settings attributes
        assert settings.directory == self.defaultdirectory
        assert settings.defaultdirectory == self.defaultdirectory 
        assert settings.filename == self.defaultfilename
        assert settings.defaultfilename == self.defaultfilename
        assert settings.list_databases == []
        assert settings.databases == {}

        # Cleanup
        if self.defaultdirectory.exists():
            shutil.rmtree(self.defaultdirectory)

    def test_databases(self):
        """Test setting and unsetting databases"""

        # Initialize a settings object
        settings = self.settings

        # Build content dict for comparison
        content = {}

        # Set a database with required terms
        settings.set_database('pytest1', 'local', 'nowhere', key='no')
        
        # Update the comparison content
        content['database'] = {}
        content['database']['pytest1'] = {'style': 'local', 'host': 'nowhere', 'key': 'no'}

        # Check status of settings attributes
        assert settings.list_databases == ['pytest1']
        assert settings.databases['pytest1'] == content['database']['pytest1']

        # Check set database with missing name
        with patch('builtins.input') as mock_input:
            mock_input.return_value = 'pytest2'
            settings.set_database(style='local', host='somewhere', key='yes')
        
        # Update the comparison content
        content['database']['pytest2'] = {'style': 'local', 'host': 'somewhere', 'key': 'yes'}
        
        # Check status of settings attributes
        assert settings.list_databases == ['pytest1', 'pytest2']
        assert settings.databases['pytest2'] == content['database']['pytest2']

        # Check set database with missing style
        with patch('builtins.input') as mock_input:
            mock_input.return_value = 'mongo'
            settings.set_database(name='pytest3', host='everywhere', key='maybe')

        # Update the comparison content
        content['database']['pytest3'] = {'style': 'mongo', 'host': 'everywhere', 'key': 'maybe'}

        # Check status of settings attributes
        assert settings.list_databases == ['pytest1', 'pytest2', "pytest3"]
        assert settings.databases['pytest3'] == content['database']['pytest3']

        # Check set database with missing host
        with patch('builtins.input') as mock_input:
            mock_input.return_value = 'elsewhere'
            settings.set_database(name='pytest4', style='cdcs', key='maybe')

        # Update the comparison content
        content['database']['pytest4'] = {'style': 'cdcs', 'host': 'elsewhere', 'key': 'maybe'}

        # Check status of settings attributes
        assert settings.list_databases == ['pytest1', 'pytest2', 'pytest3', 'pytest4']
        assert settings.databases['pytest4'] == content['database']['pytest4']

        # Check set database with missing key
        with patch('builtins.input') as mock_input:
            mock_input.return_value = ''
            settings.set_database(name='pytest5', style='cdcs', host='here')
        settings.set_database(name='pytest6', style='local', host='there', prompt=False)

        # Update the comparison content
        content['database']['pytest5'] = {'style': 'cdcs', 'host': 'here'}
        content['database']['pytest6'] = {'style': 'local', 'host': 'there'}

        # Check status of settings attributes
        assert settings.list_databases == ['pytest1', 'pytest2', 'pytest3', 'pytest4', 'pytest5', 'pytest6']
        assert settings.databases['pytest5'] == content['database']['pytest5']
        assert settings.databases['pytest6'] == content['database']['pytest6']

        # Test that prompt = False fails for missing values
        with raises(TypeError):
            settings.set_database(style='local', host='nowhere', key='no', prompt=False)
        with raises(TypeError):
            settings.set_database(name='asf', host='nowhere', key='no', prompt=False)
        with raises(TypeError):
            settings.set_database(name='asf', style='local', key='no', prompt=False)
        
        # Test updating existing settings with negative confirmation
        with patch('builtins.input') as mock_input:
            mock_input.return_value = 'no'
            settings.set_database('pytest1', 'mongo', 'nowhere', key='no')
        assert settings.databases['pytest1'] == content['database']['pytest1']
        
        # Test updating existing settings with invalid confirmation
        with raises(ValueError):
            with patch('builtins.input') as mock_input:
                mock_input.return_value = 'ag'
                settings.set_database('pytest1', 'mongo', 'nowhere', key='no')

        # Test updating existing settings with positive confirmation
        with patch('builtins.input') as mock_input:
            mock_input.return_value = 'yes'
            settings.set_database('pytest1', 'mongo', 'nowhere', key='no')
        content['database']['pytest1'] = {'style': 'mongo', 'host': 'nowhere', 'key': 'no'}
        assert settings.databases['pytest1'] == content['database']['pytest1']

        # Test updating existing settings without prompt confirmation
        settings.set_database('pytest1', 'cdcs', 'nowhere', key='no', prompt=False)
        content['database']['pytest1'] = {'style': 'cdcs', 'host': 'nowhere', 'key': 'no'}
        assert settings.databases['pytest1'] == content['database']['pytest1']

        # Test unsetting databases
        with patch('builtins.input') as mock_input:
            mock_input.return_value = 'yes'
            settings.unset_database('pytest1')
        assert settings.list_databases == ['pytest2', 'pytest3', 'pytest4', 'pytest5', 'pytest6']

        with patch('builtins.input') as mock_input:
            mock_input.return_value = 'no'
            settings.unset_database('pytest2')
        assert settings.list_databases == ['pytest2', 'pytest3', 'pytest4', 'pytest5', 'pytest6']

        settings.unset_database('pytest4', prompt=False)
        assert settings.list_databases == ['pytest2', 'pytest3', 'pytest5', 'pytest6']

        with raises(ValueError):
            settings.unset_database('badname')

        settings.unset_database('pytest2', prompt=False)
        settings.unset_database('pytest3', prompt=False)
        settings.unset_database('pytest5', prompt=False)
        settings.unset_database('pytest6', prompt=False)
        assert settings.content == {}

        # Cleanup
        if self.defaultdirectory.exists():
            shutil.rmtree(self.defaultdirectory)

    def test_directory(self, tmpdir):
        
        # Initialize a settings object
        settings = self.settings
        self.directory = tmpdir

        # Set a database for testing purposes
        settings.set_database('pytest1', 'local', 'nowhere', key='no')
        assert settings.list_databases == ['pytest1']

        # Set directory and move
        with patch('builtins.input') as mock_input:
            mock_input.return_value = 'yes'
            settings.set_directory(tmpdir, move=True)

        # Check status of settings attributes
        assert settings.directory == self.directory
        assert settings.defaultdirectory == self.defaultdirectory 
        assert settings.filename == self.filename
        assert settings.defaultfilename == self.defaultfilename
        assert settings.list_databases == ['pytest1']
        
        # Set a database for testing purposes
        settings.set_database('pytest2', 'local', 'nowhere', key='no')
        assert settings.list_databases == ['pytest1', 'pytest2']

        # unset directory and move
        with patch('builtins.input') as mock_input:
            mock_input.return_value = 'yes'
            settings.unset_directory(move=True)

        # Check status of settings attributes
        assert settings.directory == self.defaultdirectory
        assert settings.defaultdirectory == self.defaultdirectory 
        assert settings.filename == self.defaultfilename
        assert settings.defaultfilename == self.defaultfilename
        assert settings.list_databases == ['pytest1', 'pytest2']
        assert not self.filename.is_file()

        # Set directory without move
        settings.set_directory(tmpdir, move=False, prompt=False)
        
        settings.set_database('pytest3', 'local', 'nowhere', key='no')

        # Check status of settings attributes
        assert settings.directory == self.directory
        assert settings.defaultdirectory == self.defaultdirectory 
        assert settings.filename == self.filename
        assert settings.defaultfilename == self.defaultfilename
        assert settings.list_databases == ['pytest3']

        # Error for unset+move when default location contains settings 
        with raises(ValueError):
            settings.unset_directory(move=True, prompt=False)

        # Unset without move
        settings.unset_directory(move=False, prompt=False)

        # Check status of settings attributes
        assert settings.directory == self.defaultdirectory
        assert settings.defaultdirectory == self.defaultdirectory 
        assert settings.filename == self.defaultfilename
        assert settings.defaultfilename == self.defaultfilename
        assert settings.list_databases == ['pytest1', 'pytest2']

        # Error for set+move when new location contains settings
        with raises(ValueError):
            settings.set_directory(tmpdir, move=True, prompt=False)

        # Cleanup
        if self.defaultdirectory.exists():
            shutil.rmtree(self.defaultdirectory)


# coding: utf-8
__all__ = ['valuemanager', 'Value', 'load_value']

# Standard Python libraries
from typing import Any, Optional, Union

# Relative imports
#from . import str_contains, str_match, in_list, int_match, list_contains, date_match, float_match
from .Value import Value

# Initialize a ModuleManager for the query styles
from ..tools import ModuleManager
valuemanager = ModuleManager('Value')


# Add the modular Value styles
valuemanager.import_style('base', '.Value', __name__)
valuemanager.import_style('str', '.StrValue', __name__)
valuemanager.import_style('strlist', '.StrListValue', __name__)
valuemanager.import_style('longstr', '.LongStrValue', __name__)
valuemanager.import_style('float', '.FloatValue', __name__)
valuemanager.import_style('int', '.IntValue', __name__)
valuemanager.import_style('bool', '.BoolValue', __name__)
valuemanager.import_style('date', '.DateValue', __name__)
valuemanager.import_style('month', '.MonthValue', __name__)
valuemanager.import_style('floatarray', '.FloatArrayValue', __name__)

def load_value(style: str,
               name: Optional[str] = None,
               record: Optional[str] = None,
               defaultvalue: Optional[Any] = None,
               valuerequired: bool = False,
               allowedvalues: Optional[tuple] = None,
               metadatakey: Union[str, bool, None] = None,
               metadataparent: Optional[str] = None,
               modelpath: Optional[str] = None,
               description: Optional[str] = None,
               **kwargs) -> Value:
    """
    Loads a Value subclass associated with a given value style.

    Parameters
    ----------
    style : str
        The value style.
    name : str
        The name for the parameter value.  This should correspond to the name of
        the associated class attribute.
    record : Record, optional
        The Record object that the Parameter is used with.
    defaultvalue : any or None, optional
        The default value to use for the property.  The default value of
        None indicates that there is no default value.
    valuerequired: bool, optional
        Indicates if a value must be given for the property.  If True, then
        checks will be performed that a value is assigned to the property.
    allowedvalues : tuple or None, optional
        A list/tuple of values that the parameter is restricted to have.
        Setting this to None (default) indicates any value is allowed.
    metadatakey: str, bool or None, optional
        The key name to use for the property when constructing the record
        metadata dict.  If set to None (default) then name will be used for
        metadatakey.  If set to False then the parameter will not be
        included in the metadata dict.
    metadataparent: str or None, optional
        If given, then this indicates that the metadatakey is actually an
        element of a dict in metadata with this name.  This allows for limited
        support for metadata having embedded dicts.
    modelpath: str, optional
        The period-delimited path after the record root element for
        where the parameter will be found in the built data model.  If set
        to None (default) then name will be used for modelpath.
    description: str or None, optional
        A short description for the value.  If not given, then the record name
        will be used.
    **kwargs : any, optional
        Any additional style-specific keyword parameters.
    """
    return valuemanager.init(style, name=name, record=record, 
                             defaultvalue=defaultvalue,
                             valuerequired=valuerequired,
                             allowedvalues=allowedvalues,
                             metadatakey=metadatakey,
                             metadataparent=metadataparent,
                             modelpath=modelpath, 
                             description=description, **kwargs)



# coding: utf-8

# Standard Python libraries
import ast
from typing import Optional, Union

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# http://www.numpy.org/
import numpy as np
import numpy.typing as npt

class UnitConverter():

    # Import derived units defined in separate files
    from ._length import (cm, mm, um, nm, pm, fm, km,
                          angstrom, Å,
                          lightyear, astro_unit,
                          pc, kpc, Mpc, Gpc,
                          inch, foot, mile, thou)
    from ._volume import (L, mL, uL, nL, pL, fL, aL, kL, ML, GL)
    from ._time import (ms, us, ns, ps, fs,
                        minute, hour, day, week, year)
    from ._frequency import (Hz, mHz, kHz, MHz, GHz, THz, PHz, rtHz, rpm,
                             Hz·2π, mHz·2π, kHz·2π, MHz·2π, GHz·2π, THz·2π, PHz·2π, rpm·2π)
    from ._mass import (g, mg, ug, ng, pg, fg,
                        tonne, amu, Da, kDa, lbm)
    from ._energy import (J, mJ, uJ, nJ, pJ, fJ, kJ, MJ, GJ, erg,
                          eV, meV, keV, MeV, GeV, TeV,
                          btu, smallcal, kcal, Wh, kWh)
    from ._concentration import (mol, mmol, umol, nmol, pmol, fmol,
                                 M, mM, uM, nM, pM, fM)
    from ._force import (N, mN, uN, nN, pN, fN, kN, MN, GN, dyn, lbf)
    from ._pressure import (Pa, hPa, kPa, MPa, GPa, bar,
                            mbar, cbar, dbar, kbar, Mbar,
                            atm, torr, mtorr, psi)
    from ._power import (W, mW, uW, nW, pW, kW, MW, GW, TW,
                         horsepower_imperial, horsepower_metric)
    from ._acceleration import (Gal, mGal, uGal, eotvos)
    from ._temperature import (degFinterval, degCinterval, mK, uK, nK, pK)
    from ._charge import (mC, uC, nC, Ah, mAh)
    from ._current import (A, mA, uA, nA, pA, fA)
    from ._voltage import (V, mV, uV, nV, kV, MV, GV, TV)
    from ._resistance import (ohm, Ω, mohm, mΩ, kohm, kΩ, Mohm, MΩ, Gohm, GΩ,
                              S, mS, uS, nS)
    from ._magnetic import (T, mT, uT, nT, G, mG, uG, kG, Oe, Wb)
    from ._capacitance import (F, uF, nF, pF, fF, aF, H, mH, uH, nH)
    from ._constants import (NA, pi, π, c0, mu0, μ0, eps0, ε0, Z0,
                             hPlanck, hbar, ħ, kB, GNewton, sigmaSB, σSB, alphaFS, αFS,
                             Rgas, e, uBohr, uNuc, aBohr, me, mp, mn,
                             Rinf, Ry, Hartree, ARichardson, Phi0, KJos, RKlitz, debye,
                             REarth, g0, Msolar, MEarth)

    def __init__(self):
        """Initialize a unit manager and set core units"""
        
        # Init core Si units
        self.__m = 1.
        self.__kg = 1.
        self.__s = 1.
        self.__C = 1.
        self.__K = 1.

        # Init unit and unitdoc dicts
        self.__unit = {}
        self.__unitdoc = {}
        self.build_unit()

    @property
    def m(self) -> float:
        """meter"""
        return self.__m

    @property
    def kg(self) -> float:
        """kilogram"""
        return self.__kg

    @property
    def s(self) -> float:
        """second"""
        return self.__s
    
    @property
    def C(self) -> float:
        """coulomb"""
        return self.__C
    
    @property
    def K(self) -> float:
        """kelvin"""
        return self.__K

    @property
    def display_core_values(self) -> str:
        """str: the current values of the core units"""
        return '\n'.join([
            f'm  = {self.m:.15}',
            f'kg = {self.kg:.15}',
            f's  = {self.s:.15}',
            f'C  = {self.C:.15}',
            f'K  = {self.K:.15}',
        ])

    def reset_units(self, 
                    seed: Union[int, str, None] = None,
                    **kwargs):
        """
        Resets the core units from which all others are derived.  Based on the
        given inputs, three behaviors can occur:

        1. If no kwargs are given and seed is int or None, then the core units
           will be set to random values using the seed value.
        
        2. If no kwargs are given and seed is set as 'SI', then the working
           units will be set to SI units.
        
        3. If kwargs are given, then the working units will be set/derived
           based on which are given and SI units if needed.  At most 4 kwargs
           of length, mass, time, energy, and charge can be given.  And seed
           must be left as None.
        
        Parameters
        ----------
        seed : int, str or None, optional
            random number seed to use in generating random working units.
            seed='SI' will use SI units.  Must be None (default) if any other
            parameters are given.
        length : str, optional
            Unit of length to use for the working units.
        mass : str, optional
            Unit of mass to use for the working units.
        time : str, optional
            Unit of time to use for the working units.
        energy : str, optional
            Unit of energy to use for the working units.
        charge : str, optional
            Unit of charge to use for the working units.
            
        Raises
        ------
        ValueError
            If seed is given with any other parameters, or if more than four of
            the working unit parameters are given.
        """
        
        # Set known units
        if isinstance(seed, str):
            
            if len(kwargs) > 0:
                raise ValueError('seed cannot be given with any other parameters')
            
            if seed == 'SI':
                self.__m = 1.
                self.__kg = 1.
                self.__s = 1.
                self.__C = 1.
                self.__K = 1.

            else:
                raise ValueError(f'str seed value {seed} not supported')

        # Set random units
        elif len(kwargs) == 0:
            
            import random

            prior_random_state = random.getstate()

            if seed is None:
                random.seed()
            else:
                random.seed(seed)

            self.__m = 10 ** random.uniform(-2,2)
            self.__kg = 10 ** random.uniform(-2,2)
            self.__s = 10 ** random.uniform(-2,2)
            self.__C = 10 ** random.uniform(-2,2)
            self.__K = 10 ** random.uniform(-2,2)

            # Leave the random generator like I found it, in case something else is
            # using it.
            random.setstate(prior_random_state)

        elif seed is None:
            
            # Check that no more than 4 working units are defined
            if len(kwargs) > 4:
                raise ValueError('Only four working units can be defined')
            
            # Set base units to 1 (working units to SI)
            self.reset_units('SI')
            
            # Scale base units by working units
            if 'length' in kwargs:
                self.__m = self.m / self.unit[kwargs['length']]
            
            if 'mass' in kwargs:
                self.__kg = self.kg / self.unit[kwargs['mass']]
            
            if 'time' in kwargs:
                self.__s = self.s / self.unit[kwargs['time']]
            
            if 'charge' in kwargs:
                self.__C = self.C / self.unit[kwargs['charge']]
            
            # Scale derived units by working units
            if 'energy' in kwargs:
                J = self.unit['J'] / self.unit[kwargs['energy']]
                
                # Scale base units by derived units
                if 'mass' not in kwargs:
                    self.__kg = J * self.s**2 / self.m**2
                elif 'time' not in kwargs:
                    self.__s = (self.kg * self.m**2 / J)**0.5
                elif 'length' not in kwargs:
                    self.__m = (J * self.s**2 / self.kg)
            
        else:
            raise ValueError('seed cannot be given with any other parameters')
        
        # Build unit and unitdoc dicts
        self.build_unit()
        
    @property
    def unit(self) -> dict:
        """dict: All available unit names and set values"""
        return self.__unit
    
    @property
    def unitdoc(self) -> dict:
        """dict: All available unit names and their descriptions"""
        return self.__unitdoc

    def build_unit(self):
        """
        Saves numericalunits attributes to global dictionary unit so the values
        can be retrieved by their string names.
        """
        
        self.__unit = {}
        self.__unitdoc = {}
        
        # Copy all float attributes of numericalunits to unit
        for key in dir(self):
            
            value = getattr(self, key)
            
            if key[:1] != '_' and isinstance(value, float):
                self.unit[key] = value
                self.unitdoc[key] = getattr(UnitConverter, key).__doc__

    def set_literal(self, term: str) -> npt.ArrayLike:
        """
        Convert string 'value unit' to numbers in working units.
        
        Parameters
        ----------
        term : str
            String containing value and associated unit. If unit is not given,
            then the value is converted to a float and assumed to be in working
            units.
            
        Returns
        -------
        float or numpy.ndarray
            The numerical value of term in working units.
            
        Raises
        ------
        ValueError
            If no valid float value can be parsed.
        """
        
        # Set splitting point j to end of term (i.e. assume no units given)
        j = len(term)
        
        # Loop until done
        while True:
            
            # Split term into value, unit terms
            value = term[:j].strip()
            unit = term[j:].strip()
            if len(unit) == 0:
                unit = None

            # Return number if value, unit pair is valid
            try: 
                return self.set_in_units(ast.literal_eval(value), unit)
            except: 
                # Find the next splitting point
                try:
                    j = term[:j].rindex(' ')
                except ValueError as err:
                    raise ValueError('Failed to parse term') from err

    def set_in_units(self, 
                     value: npt.ArrayLike,
                     units: str) -> npt.ArrayLike:
        """
        Convert value from specified units to working units.
        
        Parameters
        ----------
        value : array-like object
            A numerical value or list/array of values.
        units : str
            The units that value is in.
            
        Returns
        -------
        float or numpy.ndarray
            The given value converted from the specified units to working units.
        """
        units = self.parse(units)
        return np.asarray(value) * units

    def get_in_units(self, 
                     value: npt.ArrayLike,
                     units: str) -> npt.ArrayLike:
        """
        Convert value from working units to specified units.
        
        Parameters
        ----------
        value : array-like object
            A numerical value or list/array of values.
        units : str
            The units to convert value to (from working units).
            
        Returns
        -------
        float or numpy.ndarray
            The given value converted to the specified units from working units.
        """
        units = self.parse(units)
        return np.asarray(value) / units

    def value_unit(self, term: dict) -> npt.ArrayLike:
        """
        Reads numerical value from dictionary containing 'value' and 'unit' keys.
        
        Parameters
        ----------
        term : dict
            Dictionary containing 'value' and 'unit' keys.
            
        Returns
        -------
        float or numpy.ndarray
            The result of calling set_in_units() by passing the dictionary keys 
            'value' and 'unit' as parameters.
        
        """
        unit = term.get('unit', None)
        if unit is None:
            value = np.asarray(term['value'])
        else:
            value = self.set_in_units(term['value'], unit)
        
        if 'shape' in term:
            shape = tuple(term['shape'])
            value = value.reshape(shape)
        
        return value
        
    def error_unit(self, term: dict) -> npt.ArrayLike:
        """
        Reads numerical error from dictionary containing 'error' and 'unit' keys.
        
        Parameters
        ----------
        term : dict
            Dictionary containing 'error' and 'unit' keys.
            
        Returns
        -------
        float or numpy.ndarray
            The result of calling set_in_units() by passing the dictionary keys 
            'error' and 'unit' as parameters.
        
        """
        unit = term.get('unit', None)
        if unit is None:
            error = np.asarray(term['error'])
        else:
            error = self.set_in_units(term['error'], unit)
        
        if 'shape' in term:
            shape = tuple(term['shape'])
            error = error.reshape(shape)
        
        return error
        
    def model(self,
              value: npt.ArrayLike,
              units: Optional[str] = None,
              error: Optional[npt.ArrayLike] = None) -> DM:
        """
        Generates DataModelDict representation of data.
        
        Parameters
        ----------
        value : array-like object
            A numerical value or list/array of values.
        units : str, optional
            The units to convert value to (from working units).
        error : array-like object or None, optional
            A value error to include.  If given, must be the same
            size/shape as value.
        
        Returns
        -------
        DataModelDict
            Model representation of the value(s).
        """
        
        datamodel = DM()
        
        if units is not None:
            value = self.get_in_units(value, units)
        
        if error is not None:
            error = self.get_in_units(error, units)
        
        # Single value
        if value.ndim == 0:
            datamodel['value'] = value
            if error is not None:
                datamodel['error'] = error
        
        # 1D array
        elif value.ndim == 1:
            datamodel['value'] = value.tolist()
            if error is not None:
                datamodel['error'] = error.tolist()
                
        # Higher-order array requires shape
        else:
            shape = value.shape
            datamodel['value'] = value.flatten().tolist()
            if error is not None:
                datamodel['error'] = error.flatten().tolist()
            datamodel['shape'] = list(shape)
        
        if units is not None:
            datamodel['unit'] = units
        
        return datamodel

    def parse(self, units: Optional[str]) -> float:
        """
        Convert units as strings (or None) into scaling numbers.  This function
        allows for complex unit definitions with operators:
        
        - '()' for defining order of operations
        - '*' for multiplication.
        - '/' for division.
        - '^' for powers.
        
        Parameters
        ----------
        units : str or None
            String consisting of defined unit names, operators, and numerical 
            values to interpret.
            
        Returns
        -------
        float
            The scaling factor for converting numbers in the given units to
            working units. If units is None or == 'scaled', then this value is
            1.0.
        """
        
        # Units of None does no scaling
        if units is None or units == 'scaled':
            return 1

        # Parse string and return number value
        elif isinstance(units, str):
            i = 0
            terms = []
            
            # Break into terms
            while i < len(units):
                
                # parse terms in parentheses first
                if units[i] == '(':
                    j = i + 1
                    pcount = 0
                    while True:
                        if j == len(units):
                            raise ValueError('Invalid () terms.')
                        elif units[j] == ')':
                            if pcount == 0:
                                break
                            else:
                                pcount -= 1
                        elif units[j] == '(':
                            pcount += 1
                        j += 1
                        
                    terms.append(self.parse(units[i+1: j]))
                    i = j + 1
                
                # append string terms
                elif units[i].isalpha():
                    term = ''
                    while i < len(units) and units[i] not in ' */^\n\r\t':
                        term += units[i]
                        i += 1
                    terms.append(self.unit[term])
                
                # append numeric terms
                elif units[i].isdigit() or units[i] == '-' or units[i] == '.':
                    term = ''
                    while i < len(units) and units[i] not in ' */^\n\r\t':
                        term += units[i]
                        i += 1
                    terms.append(float(term))
                
                # append operators
                elif units[i] in '*/^':
                    terms.append(units[i])
                    i += 1
                
                # ignore excess white characters
                elif units[i] in ' \n\r\t':
                    i += 1
                
                # issue error for unmatched ) parentheses
                elif units[i] == ')':
                    raise ValueError('Invalid () terms.')
                
                else:
                    raise ValueError('Unknown character: %s' % units[i])
            
            # Compute powers
            while '^' in terms:
                c = terms.index('^')
                value = [terms[c-1] ** terms[c+1]]
                terms = terms[:c-1] + value + terms[c+2:]
            
            # Compute multiplication and division
            while len(terms) > 1:
                if terms[1] == '*':
                    value = [terms[0] * terms[2]]
                    terms = value + terms[3:]
                elif terms[1] == '/':
                    value = [terms[0] / terms[2]]
                    terms = value + terms[3:]
                else:
                    raise ValueError('Invalid string format')
            
            return terms[0]

        # Else assume units is already a number
        else:
            return units
        
unitconvert = UnitConverter()


# coding: utf-8
# Standard Python libraries
from pathlib import Path
from importlib import resources
from typing import Union, Optional, Any
import io
from tarfile import TarFile

# https://ipython.org/
from IPython.display import display, HTML

from PIL import Image

# https://lxml.de/
import lxml.etree as ET

import pandas as pd

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

from .. import load_query, load_value

class Record():
    """
    Class for handling different record styles in the same fashion.  The
    base class defines the common methods and attributes.
    """

    def __init__(self,
                 model: Union[str, io.IOBase, DM, None] = None,
                 name: Optional[str] = None,
                 database = None,
                 noname: bool = False,
                 **kwargs: any):
        """
        Initializes a Record object for a given style.
        
        Parameters
        ----------
        model : str, file-like object, or DataModelDict, optional
            The contents of the record.
        name : str, optional
            The unique name to assign to the record.  If model is a file
            path, then the default record name is the file name without
            extension.
        database : yabadaba.Database, optional
            A default Database to associate with the Record, typically the
            Database that the Record was obtained from.  Can allow for Record
            methods to perform Database operations without needing to specify
            which Database to use.
        nonoame : bool, optional
            Flag indicating if the record does not get assigned a name.  This
            is primarily for when a record class is used as a value inside
            another class, i.e., this record is a subset of another.
        kwargs : any
            Any record-specific attributes to assign.
        """
        self.__model = None
        self.__name = None
        self.tar = None
        self.database = database
        self.noname = noname

        self.__value_objects = None
        self.__value_objects = tuple(self._init_value_objects())
        self._init_values()
        if len(self.__value_dict) > 0:
            if len(self.value_objects) > 0:
                raise ValueError('Value objects should not be set using both _init_value_objects and _init_value_dict')
            self.__value_objects = tuple(self.__value_dict.values())

        if model is not None:
            assert len(kwargs) == 0, f"cannot specify kwargs with model: '{kwargs.keys()}'"
            self.load_model(model, name=name)
        else:
            self.set_values(name=name, **kwargs)

    def _init_value_objects(self) -> list:
        """
        Method that defines the value objects for the Record.  This should
        1. Call the method's super().
        2. Use yabadaba.load_value() to build Value objects that are set to
           private attributes of self.
        3. Append the list returned by the super() with the new Value objects.

        Returns
        -------
        value_objects: A list of all value objects.
        """
        if self.__value_objects is not None:
            raise RuntimeError('_init_value_objects should only be called by Record.__init__')

        return []

    def _init_values(self):
        """
        Method that defines the value objects for the Record.  This should
        call the super of this method, then use self._add_value to create new Value objects.
        Note that the order values are defined matters
        when build_model is called!!!
        """
        pass

    def _add_value(self,
                   style: str,
                   name: str,
                   defaultvalue: Optional[Any] = None,
                   valuerequired: bool = False,
                   allowedvalues: Optional[tuple] = None,
                   metadatakey: Union[str, bool, None] = None,
                   metadataparent: Optional[str] = None,
                   modelpath: Optional[str] = None,
                   description: Optional[str] = None,
                   **kwargs):
        """
        Method to add Value objects when defining the class.  This should only be
        used when defining _init_values()!

        Parameters
        ----------
        style : str
            The value style.
        name : str
            The name for the parameter value.  This corresponds to the name of
            the associated class attribute.
        defaultvalue : any or None, optional
            The default value to use for the property.  The default value of
            None indicates that there is no default value.
        valuerequired: bool, optional
            Indicates if a value must be given for the property.  If True, then
            checks will be performed that a value is assigned to the property.
        allowedvalues : tuple or None, optional
            A list/tuple of values that the parameter is restricted to have.
            Setting this to None (default) indicates any value is allowed.
        metadatakey: str, bool or None, optional
            The key name to use for the property when constructing the record
            metadata dict.  If set to None (default) then name will be used for
            metadatakey.  If set to False then the parameter will not be
            included in the metadata dict.
        metadataparent: str or None, optional
            If given, then this indicates that the metadatakey is actually an
            element of a dict in metadata with this name.  This allows for limited
            support for metadata having embedded dicts.
        modelpath: str, optional
            The period-delimited path after the record root element for
            where the parameter will be found in the built data model.  If set
            to None (default) then name will be used for modelpath.
        description: str or None, optional
            A short description for the value.  If not given, then the record name
            will be used.
        **kwargs : any, optional
            Any additional style-specific keyword parameters.
        """
        self.__value_dict[name] = load_value(style=style, name=name, record=self,
                                             defaultvalue=defaultvalue,
                                             valuerequired=valuerequired,
                                             allowedvalues=allowedvalues,
                                             metadatakey=metadatakey,
                                             metadataparent=metadataparent,
                                             modelpath=modelpath,
                                             description=description,
                                             **kwargs)

    def __getattr__(self, name: str):
        """Adjusted to get value attributes from Value objects"""
        if name.startswith('__'):
            super().__getattr__(self, name)

        elif name == '_Record__value_dict':
            self.__value_dict = {}
            return self.__value_dict
        
        elif name in self.__value_dict:
            return self.__value_dict[name].value
    
    def __setattr__(self, name: str, value: Any):
        """Adjusted to set to value attributes of Value objects"""
        if name == '_Record__value_dict' or name.startswith('__'):
            super().__setattr__(name, value)
        elif name in self.__value_dict:
            self.__value_dict[name].value = value
        else:
            super().__setattr__(name, value)


    def load_model(self,
                   model: Union[str, io.IOBase, DM],
                   name: Optional[str] = None):
        """
        Loads record contents from a given model.

        Parameters
        ----------
        model : str, file-like object, or DataModelDict
            The model contents of the record to load.
        name : str, optional
            The name to assign to the record.  Often inferred from other
            attributes if not given.
        """
        # Get name if model is a filename
        if name is None:
            try:
                if Path(model).is_file():
                    self.name = Path(model).stem
            except (ValueError, OSError, TypeError):
                pass
        else:
            self.name = name

        # Read/set model
        self._set_model(model)

        # Extract parameter values 
        rec = self.model[self.modelroot]
        for value_object in self.value_objects:
            value_object.load_model(rec)

        if self.noname is False:
            try:
                assert self.name is not None
            except:
                try:
                    self.name = self.defaultname
                except:
                    self.name = None

    def set_values(self, **kwargs):
        """
        Set multiple object attributes at the same time.

        Parameters
        ----------
        **kwargs: any
            Any parameters for the record that you wish to set values for.
        """
        if 'name' in kwargs:
            self.name = kwargs['name']
        
        for value_object in self.value_objects:
            if value_object.name in kwargs:
                setattr(self, value_object.name, kwargs[value_object.name])

        if self.noname is False:
            try:
                assert self.name is not None
            except:
                try:
                    self.name = self.defaultname
                except:
                    self.name= None

    def __str__(self) -> str:
        """str: The string representation of the record"""
        return f'{self.style} record named {self.name}'

    @property
    def style(self) -> str:
        """str: The record style"""
        return 'base'

    @property
    def xsd_filename(self) -> tuple:
        """tuple: The module path and file name of the record's xsd schema"""
        raise NotImplementedError('Not implemented')

    @property
    def xsd(self) -> bytes:
        """bytes: The xml schema for the record style."""
        return resources.read_binary(*self.xsd_filename)

    @property
    def xsl_filename(self) -> tuple:
        """tuple: The module path and file name of the record's xsl html transformer"""
        raise NotImplementedError('Not implemented')

    @property
    def xsl(self) -> bytes:
        """bytes: The xsl transformer for the record style."""
        return resources.read_binary(*self.xsl_filename)

    @property
    def name(self) -> str:
        """str: The record's name."""
        if self.__name is not None or self.noname:
            return self.__name
        else:
            raise AttributeError('record name not set')

    @name.setter
    def name(self, value: Optional[str]):
        if value is not None:
            if self.noname:
                raise TypeError('name turned off for this record')
            self.__name = str(value)
        else:
            self.__name = None

    @property
    def defaultname(self) -> Optional[str]:
        """str: The name to default to, usually based on other properties"""
        return None

    @property
    def noname(self) -> bool:
        """bool: Indicates that the record should not have a name."""
        return self.__noname

    @noname.setter
    def noname(self, val: bool):
        if not isinstance(val, bool):
            raise TypeError('noname must be a bool')
        self.__noname = val

    @property
    def value_objects(self) -> tuple:
        """tuple: The Value objects associated with the Record"""
        return self.__value_objects

    def get_value(self, name):
        """Returns a Value object from __value_dict by name"""
        return self.__value_dict[name]

    @property
    def modelroot(self) -> str:
        """str : The name of the root element in the model contents."""
        raise NotImplementedError('Specific to subclasses')

    @property
    def model(self) -> DM:
        """DataModelDict: The record's model content."""
        if self.__model is not None:
            return self.__model
        else:
            raise AttributeError('model content has not been loaded or built')

    def reload_model(self):
        """
        Reloads the record based on the model content.  This allows for direct
        changes to the model to be updated to the object. 
        """
        self.load_model(model=self.model, name=self.name)

    def _set_model(self, model: DM):
        """
        Sets model content - called by build_model() and load_model() to update
        content.  Use load_model() if you are passing in an external model.
        """
        try:
            modelroot = self.modelroot
        except NotImplementedError:
            self.__model = DM(model)
        else:
            # Load model as DataModelDict
            content = DM(model).find(modelroot)
            self.__model = DM([(modelroot, content)])

    def build_model(self):
        """
        Generates and returns model content based on the values set to object.
        """
        self.__model = DM()
        self.__model[self.modelroot] = rec = DM()

        for value_object in self.value_objects:
            value_object.build_model(rec)        

        return self.__model

    def metadata(self) -> dict:
        """
        Generates a dict of simple metadata values associated with the record.
        Useful for quickly comparing records and for building pandas.DataFrames
        for multiple records of the same style.
        """
        meta = {}
        # Initialize with only name
        if self.noname is False:
            meta['name'] = self.name

        # Add value object values
        for value_object in self.value_objects:
            value_object.metadata(meta)
        
        return meta

    @property
    def metadatakeys(self) -> list:
        """list: The keys included in the metadata dict"""
        keys = []
        if self.noname is False:
            keys.append('name')
        
        # Add value object values
        for value_object in self.value_objects:
            if value_object.metadataparent is not None:
                keys.append(value_object.metadataparent)
            elif value_object.metadatakey is not False:
                keys.append(value_object.metadatakey)
        
        return keys

    @property
    def queries(self) -> dict:
        """dict: Query objects and their associated parameter names."""
        
        # Build dict containing all queries of all values
        queries = {}
        for value_object in self.value_objects:
            queries.update(value_object.queries)
        
        return queries

    @property
    def querynames(self) -> list:
        """list: The query parameter names supported by the record."""
        return list(self.queries.keys())

    @property
    def querydoc(self) -> str:
        """str: A description of all the queries supported by the record."""
        doc = f'# {self.style} Query Parameters\n\n'
        for name, query in self.queries.items():
            doc += f'- __{name}__ (*{query.parameter_type}*): {query.description}\n'

        return doc

    def pandasfilter(self,
                     dataframe: pd.DataFrame,
                     name: Union[str, list, None] = None,
                     **kwargs: any) -> pd.Series:
        """
        Filters a pandas.DataFrame based on kwargs values for the record style.
        
        Parameters
        ----------
        dataframe : pandas.DataFrame
            A table of metadata for multiple records of the record style.
        name : str or list, optional
            The record name(s) to parse by.
        **kwargs : any
            Any of the record style-specific search parameters.

        Returns
        -------
        pandas.Series
            Boolean map of matching values
        """
        # Get defined queries
        queries = self.queries

        # Query name
        if self.noname is False:
            matches = load_query('str_match', name='name').pandas(dataframe, name)
        elif name is not None:
            raise ValueError('name turned off for record')

        # Apply queries based on given kwargs
        for key in kwargs:
            matches = (matches & queries[key].pandas(dataframe, kwargs[key]))

        return matches

    def mongoquery(self,
                   name: Union[str, list, None] = None,
                   **kwargs: any) -> dict:
        """
        Builds a Mongo-style query based on kwargs values for the record style.

        Parameters
        ----------
        name : str or list, optional
            The record name(s) to parse by.
        **kwargs : any
            Any of the record style-specific search parameters.

        Returns
        -------
        dict
            The Mongo-style query
        """
        # Get the dict of queries
        queries = self.queries

        # Initialize the full query dict and list of query operations
        querydict = {}
        querydict['$and'] = querylist = [{}]

        # Query name
        if self.noname is False:
            load_query('str_match', path='name').mongo(querylist, name)
        elif name is not None:
            raise ValueError('name turned off for record')

        # Apply queries based on given kwargs
        for key in kwargs:
            queries[key].mongo(querylist, kwargs[key], prefix='content.')

        return querydict

    def cdcsquery(self,
                  **kwargs: any) -> dict:
        """
        Builds a CDCS-style query based on kwargs values for the record style.
        
        Parameters
        ----------
        **kwargs : any
            Any of the record style-specific search parameters.
        
        Returns
        -------
        dict
            The CDCS-style query
        """
        # Get the dict of queries
        queries = self.queries

        # Initialize the query dictionary
        querydict = {}
        querydict['$and'] = querylist = [{}]

        # Apply queries based on given kwargs
        for key in kwargs:
            queries[key].mongo(querylist, kwargs[key])

        return querydict

    def html(self,
             render: bool = False) -> Optional[str]:
        """
        Returns an HTML representation of the object.
        
        Parameters
        ----------
        render : bool, optional
            If True, then IPython is used to render the HTML.  If False
            (default), then the HTML code is returned as a str.

        Returns
        -------
        str
            The HTML code contents.  Returned if render=False.
        """

        # Build xml content
        xml_content = self.model.xml()

        xml = ET.fromstring(xml_content.encode('UTF-8'))

        # Read xsl content
        xsl = ET.fromstring(self.xsl)

        # Transform to html
        transform = ET.XSLT(xsl)
        html = transform(xml)
        html_content = ET.tostring(html).decode('UTF-8')

        if render:
            display(HTML(html_content))
        else:
            return html_content

    def valid_xml(self,
                  xml_content: Optional[str] = None) -> bool:
        """
        Tests if XML content is valid with schema.
        
        Parameters
        ----------
        xml_content : str, optional
            XML content to test against the record's schema.
            If not given, will generate the xml using build_model.
        
        Returns
        -------
        bool
            Indicating if XML is valid.
        """

        # Build xml content
        if xml_content is None:
            xml_content = self.model.xml()

        xml = ET.fromstring(xml_content.encode('UTF-8'))

        # Read xsd content
        xsd = ET.fromstring(self.xsd)

        schema = ET.XMLSchema(xsd)
        return schema.validate(xml)

    @property
    def database(self):
        """yabadaba.Database or None: The default Database associated with the Record"""
        return self.__database

    @database.setter
    def database(self, value):
        if value is None or hasattr(value, 'get_records'):
            self.__database = value
        else:
            raise TypeError('database must be a yabadaba.Database or None')

    @property
    def tar(self):
        """tarfile.TarFile: The tar archive associated with the record"""
        # Return tarfile if set
        if self.__tar is not None:
            return self.__tar

        # Check if database is set
        if self.database is None:
            raise ValueError('tar not loaded and no database set')

        # Fetch tar from database, set to cache and return
        self.tar = self.database.get_tar(record=self)
        return self.__tar

    @tar.setter
    def tar(self, value: Optional[TarFile]):
        if value is None or isinstance(value, TarFile):
            self.__tar = value
        else:
            raise TypeError('tar must ne a TarFile or None')

    def clear_tar(self):
        """Closes and unsets the record's tar file to save memory"""
        if self.__tar is not None:
            self.__tar.close()
            self.tar = None

    def get_file(self,
                 filename: Union[str, Path],
                 localroot: Union[str, Path, None] = None,
                 local: bool = True):
        """
        Retrieves a file either locally or from the record's tar archive.

        Parameters
        ----------
        filename : str or Path
            The name/path for the file.  For local files, this is taken
            relative to localroot.  For files in the tar archive, this is taken
            relative to the tar's root directory which is always named for the
            record, i.e., {self.name}/{filename}.
        localroot : str, Path or None, optional
            The local root directory that filename (if it exists) is relative
            to.  The default value of None will use the current working
            directory.
        local : bool, optional
            If True (default) then the localroot will be checked for the file
            prior to retrieving the file from the tar.  If False then only the
            tar will be checked.
        
        Raises
        ------
        ValueError
            If filename exists in the tar but is not a file.

        Returns
        -------
        io.IOBase
            A file-like object in binary read mode that allows for the file
            contents to be read.
        """
        if local:
            # Set default root path
            if localroot is None:
                localroot = Path.cwd()
            else:
                localroot = Path(localroot)

            # Return local copy of file if it exists
            localfile = Path(localroot, filename)
            if Path(localfile).is_file():
                return open(localfile, 'rb')

        # Return file extracted from tar
        fileio = self.tar.extractfile(f'{self.name}/{filename}')
        if fileio is not None:
            return fileio
        else:
            raise ValueError(f'{filename} exists in tar, but is not a file')

    def display_image(self,
                      filename: Union[str, Path],
                      localroot: Union[str, Path, None] = None,
                      width: Optional[int] = None,
                      height: Optional[int] = None):
        """
        For IPython-based environments, this retrieves an image file either
        locally or from within the record's tar archive and displays it.

        Parameters
        ----------
        filename : str or Path
            The name/path for the file.  For local files, this is taken
            relative to localroot.  For files in the tar archive, this is taken
            relative to the tar's root directory which is always named for the
            record, i.e., {self.name}/{filename}.
        localroot : str, Path or None, optional
            The local root directory that filename (if it exists) is relative
            to.  The default value of None will use the current working
            directory.
        
        Raises
        ------
        ValueError
            If filename exists in the tar but is not a file.
        """
        fileio = self.get_file(filename=filename, localroot=localroot)
        img = Image.open(fileio)

        if width is not None:
            if height is None:
                oldwidth, oldheight = img.size
                height = round(oldheight * (width / oldwidth))
            size = (width, height)
            img = img.resize(size)

        elif height is not None:
            oldwidth, oldheight = img.size
            width = round(oldwidth * (height / oldheight))
            size = (width, height)
            img = img.resize(size)

        display(img)


@property
def mC(self) -> float:
    """millicoulomb"""
    return 1e-3 * self.C

@property
def uC(self) -> float:
    """microcoulomb"""
    return 1e-6 * self.C

@property
def nC(self) -> float:
    """nanocoulomb"""
    return 1e-9 * self.C

@property
def Ah(self) -> float:
    """amp-hour"""
    return 3600. * self.C

@property
def mAh(self) -> float:
    """milliamp-hour"""
    return 1e-3 * self.Ah


from typing import Optional, Union, Any

from DataModelDict import DataModelDict as DM

import numpy as np

from ..query import load_query
from . import Value

class MonthValue(Value):
    """Variation of IntValue focused purely on months.  Note model uses xsd:gMonth format"""
    
    @staticmethod
    def str_to_number(val: str):
        """
        Convert a str version of the month to an integer version
        """
        converter = {
            # Short month names
            'jan':1, 'feb':2, 'mar':3, 'apr':4, 'may':5, 'jun':6,
            'jul':7, 'aug':8, 'sep':9, 'oct':10, 'nov':11, 'dec':12,

            # Long month names
            'january':1, 'february':2, 'march':3, 'april':4, 'may':5, 'june':6,
            'july':7, 'august':8, 'september':9, 'october':10, 'november':11, 'december':12,
        }

        # Convert xsd:gMonth str format to Month
        if val[:2] == '--':
            return int(val[2:4])
        
        # Convert string
        elif val.lower() in converter:
            return converter[val.lower()]
        
        else:
            return int(val)
        
    @property
    def fullname(self) -> str:
        """str: The full month name"""
        if self.value is None:
            return None
        
        converter = {
            1:'January', 2:'February', 3:'March', 4:'April', 5:'May', 6:'June',
            7:'July', 8:'August', 9:'September', 10:'October', 11:'November', 12:'December'
        }
        return converter[self.value]
    
    @property
    def shortname(self) -> str:
        """str: The 3 letter short name for the month"""
        if self.value is None:
            return None
        
        converter = {
            1:'jan', 2:'feb', 3:'mar', 4:'apr', 5:'may', 6:'jun',
            7:'jul', 8:'aug', 9:'sep', 10:'oct', 11:'nov', 12:'dec'
        }
        return converter[self.value]
    
    @property
    def xsd_gmonth(self) -> str:
        """str: the xsd:gMonth format of the month"""
        if self.value is None:
            return None
        
        return f'--{self.value:02}'

    def set_value_mod(self, val):

        # Check if value is in #text
        val = self.set_value_mod_textfield(val)

        if val is None:
            return None
        
        if isinstance(val, float):
            rval = round(val)
            if not np.isclose(rval, val):
                raise TypeError(f'{self.name} not an int!')
            val = rval
        
        elif isinstance(val, str):
            val = self.str_to_number(val)
        
        elif not isinstance(val, int):
            raise TypeError('unsupported type for month value')
        
        if val < 1 or val > 12:
            raise ValueError('month value must range from 1 to 12')
        
        return val
    
    def build_model_value(self):
        """Function to modify how values are represented in the model"""
        return self.xsd_gmonth

    @property
    def _default_queries(self) -> dict:
        """dict: Default query operations to associate with the Parameter style"""
        if self.metadatakey is False:
            return {}
        else:
            return {
                self.name: load_query('month_match',
                                    name=self.metadatakey,
                                    parent=self.metadataparent,
                                    path=f'{self.record.modelroot}.{self.modelpath}',
                                    description=f'Return only the records where {self.description} matches a given month integer value')
            }


@property
def A(self) -> float:
    """ampere"""
    return self.C / self.s

@property
def mA(self) -> float:
    """milliampere"""
    return 1e-3 * self.A

@property
def uA(self) -> float:
    """microampere"""
    return 1e-6 * self.A

@property
def nA(self) -> float:
    """nanoampere"""
    return 1e-9 * self.A

@property
def pA(self) -> float:
    """picoampere"""
    return 1e-12 * self.A

@property
def fA(self) -> float:
    """femtoampere"""
    return 1e-15 * self.A


# coding: utf-8
# Standard Python libraries
from pathlib import Path
from typing import Optional, Tuple, Union
import tarfile

from tqdm import tqdm

import pandas as pd

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# iprPy imports
from ..record import recordmanager, load_record, Record
from ..tools import screen_input

class Database():
    """
    Class for handling different database styles in the same fashion.  This
    base class defines the common methods and attributes.
    """

    def __init__(self,
                 host: str):
        """
        Initializes a connection to a database.
        
        Parameters
        ----------
        host : str
            The host name (path, url, etc.) for the database.
        """
        # Check that object is a subclass
        if self.__module__ == __name__:
            raise TypeError("Don't use Database itself, only use derived classes")

        # Set property values
        self.__host = host

    def __str__(self) -> str:
        """
        Returns
        -------
        str
            The string representation of the database.
        """
        return f'database style {self.style} at {self.host}'

    @property
    def style(self) -> str:
        """str: The database style"""
        raise NotImplementedError('Not defined for base class')

    @property
    def host(self) -> str:
        """str: The database's host."""
        return self.__host

    def get_records(self, 
                    style: Optional[str] = None,
                    return_df: bool = False,
                    **kwargs) -> Union[list, Tuple[list, pd.DataFrame]]:
        """
        Produces a list of all matching records in the database.
        
        Parameters
        ----------
        style : str, optional
            The record style to search. If not given, a prompt will ask for it.
        return_df : bool, optional
            If True, then the corresponding pandas.Dataframe of metadata
            will also be returned
        **kwargs : any, optional
            Any extra options specific to the database style or metadata search
            parameters specific to the record style.
            
        Returns
        ------
        records : numpy.NDArray
            All records from the database matching the given parameters.
        records_df : pandas.DataFrame
            The corresponding metadata values for the records.  Only returned
            if return_df is True.
        
        Raises
        ------
        AttributeError
            If get_records is not defined for database style.
        """
        raise AttributeError('get_records not defined for Database style')

    def get_record(self,
                   style: Optional[str] = None,
                   **kwargs) -> Record:
        """
        Returns a single matching record from the database.  Issues an error
        if multiple or no matching records are found.
        
        Parameters
        ----------
        style : str, optional
            The record style to limit the search by.
        **kwargs : any, optional
            Any extra options specific to the database style or metadata search
            parameters specific to the record style.

        Returns
        ------
        Record
            The single record from the database matching the given parameters.
        
        Raises
        ------
        AttributeError
            If get_record is not defined for database style.
        """
        raise AttributeError('get_record not defined for Database style')

    def get_records_df(self,
                       style: Optional[str] = None,
                       **kwargs) -> pd.DataFrame:
        """
        Produces a pandas.DataFrame of all matching records in the database.
        
        Parameters
        ----------
        style : str
            The record style to collect records of.
        **kwargs : any, optional
            Any extra options specific to the database style or metadata search
            parameters specific to the record style.

        Returns
        ------
        pandas.DataFrame
            All records from the database of the given record style.
        
        Raises
        ------
        AttributeError
            If get_records_df is not defined for database style.
        """
        raise AttributeError('get_records_df not defined for Database style')

    def count_records(self,
                      style: Optional[str] = None,
                      **kwargs) -> int:
        """
        Retrieves a count of matching records from the database.  Depending on the
        database style, this may be much faster than get_records if you only want
        to know the number of records
        
        Parameters
        ----------
        style : str
            The record style to collect records of.
        **kwargs : any, optional
            Any extra options specific to the database style or metadata search
            parameters specific to the record style.

        Returns
        ------
        int
            A count of the number of matching records in the database
        
        Raises
        ------
        AttributeError
            If count_records is not defined for database style.
        """
        raise AttributeError('count_records not defined for Database style')

    def retrieve_record(self,
                        style: Optional[str] = None,
                        dest: Optional[Path] = None,
                        format: str = 'json',
                        indent: Optional[int] = 4,
                        verbose: bool = False,
                        **kwargs):
        """
        Gets a single matching record from the database and saves it to a
        file based on the record's name.

        Parameters
        ----------
        style : str, optional
            The record style to search. If not given, a prompt will ask for it.
        dest : path, optional
            The parent directory where the record will be saved to.  If not given,
            will use the current working directory.
        format : str, optional
            The file format to save the record in: 'json' or 'xml'.  Default
            is 'json'.
        indent : int or None, optional
            The number of space indentation spacings to use in the saved
            record for the different tiered levels.  Default is 4.  Giving None
            will create a compact record.
        verbose : bool, optional
            If True, info messages will be printed during operations.  Default
            value is False.
        **kwargs : any, optional
            Any extra options specific to the database style or metadata search
            parameters specific to the record style.
        """
        # Set default dest
        if dest is None:
            dest = Path.cwd()

        # Get the record
        record = self.get_record(style=style, **kwargs)

        # Save as json
        if format == 'json':
            fname = Path(dest, f'{record.name}.json')
            with open(fname, 'w', encoding='UTF-8') as f:
                record.model.json(fp=f, indent=indent, ensure_ascii=False)
            if verbose:
                print(f'{fname} saved')

        # Save as xml
        elif format == 'xml':
            fname = Path(dest, f'{record.name}.xml')
            with open(fname, 'w', encoding='UTF-8') as f:
                record.model.xml(fp=f, indent=indent)
            if verbose:
                print(f'{fname} saved')

        else:
            raise ValueError('Invalid format: must be json or xml.')

    def add_record(self,
                   record: Optional[Record] = None,
                   style: Optional[str] = None,
                   name: Optional[str] = None,
                   model: Union[str, DM, None] = None,
                   build: bool = False,
                   verbose: bool = False) -> Record:
        """
        Adds a new record to the database.
        
        Parameters
        ----------
        record : Record, optional
            The new record to add to the database.  If not given, then name,
            style and content are required.
        style : str, optional
            The record style for the new record.  Required if record is not
            given.
        name : str, optional
            The name to assign to the new record.  Required if record is not
            given.
        model : str or DataModelDict, optional
            The model contents of the new record.  Required if record is not
            given.
        build : bool, optional
            If True, then the uploaded content will be (re)built based on the
            record's attributes.  If False (default), then record's existing
            content will be loaded if it exists, or built if it doesn't exist.
        verbose : bool, optional
            If True, info messages will be printed during operations.  Default
            value is False.
            
        Returns
        ------
        Record
            Either the given record or a record composed of the name, style,
            and content.
        
        Raises
        ------
        AttributeError
            If add_record is not defined for database style.
        """
        raise AttributeError('add_record not defined for Database style')

    def update_record(self,
                      record: Optional[Record] = None,
                      style: Optional[str] = None,
                      name: Optional[str] = None,
                      model: Union[str, DM, None] = None,
                      build: bool = False,
                      verbose: bool = False) -> Record:
        """
        Replaces an existing record with a new record of matching name and
        style, but new content.
        
        Parameters
        ----------
        record : Record, optional
            The record with new content to update in the database.  If not
            given, content is required along with name and/or style to
            uniquely define a record to update.
        style : str, optional
            The style of the record to update.
        name : str, optional
            The name to uniquely identify the record to update.
        model : str or DataModelDict, optional
            The model contents of the new record.  Required if record is not
            given.
        build : bool, optional
            If True, then the uploaded content will be (re)built based on the
            record's attributes.  If False (default), then record's existing
            content will be loaded if it exists, or built if it doesn't exist.
        verbose : bool, optional
            If True, info messages will be printed during operations.  Default
            value is False.

        Returns
        ------
        Record
            Either the given record or a record composed of the name, style,
            and content.
        
        Raises
        ------
        AttributeError
            If update_record is not defined for database style.
        """
        raise AttributeError('update_record not defined for Database style')

    def delete_record(self,
                      record: Optional[Record] = None,
                      style: Optional[str] = None,
                      name: Optional[str] = None,
                      verbose: bool = False):
        """
        Permanently deletes a record from the database.
        
        Parameters
        ----------
        record : Record, optional
            The record to delete from the database.  If not given, name and/or
            style are needed to uniquely define the record to delete.
        style : str, optional
            The style of the record to delete.
        name : str, optional
            The name of the record to delete.
        verbose : bool, optional
            If True, info messages will be printed during operations.  Default
            value is False.
        
        Raises
        ------
        AttributeError
            If delete_record is not defined for database style.
        """
        raise AttributeError('delete_record not defined for Database style')

    def get_tar(self,
                record: Optional[Record] = None,
                style: Optional[str] = None,
                name: Optional[str] = None,
                raw: bool = False) -> Union[tarfile.TarFile, bytes]:
        """
        Retrieves the tar archive associated with a record in the database.
        
        Parameters
        ----------
        record : Record, optional
            The record to retrieve the associated tar archive for.
        style : str, optional
            The style to use in uniquely identifying the record.
        name : str, optional
            The name to use in uniquely identifying the record.
        raw : bool, optional
            If True, return the archive as raw binary content. If
            False, return as an open tarfile. (Default is False)
            
        Returns
        -------
        tarfile or str
            The tar archive as an open tarfile if raw=False, or as a binary
            str if raw=True.
        
        Raises
        ------
        AttributeError
            If get_tar is not defined for database style.
        """
        raise AttributeError('get_tar not defined for Database style')

    def add_tar(self, 
                record: Optional[Record] = None,
                style: Optional[str] = None,
                name: Optional[str] = None,
                tar: Optional[bytes] = None,
                root_dir: Optional[Path] = None):
        """
        Archives and stores a folder associated with a record.
        
        Parameters
        ----------
        record : Record, optional
            The record to associate the tar archive with.  If not given, then
            name and/or style necessary to uniquely identify the record are
            needed.
        style : str, optional
            The style to use in uniquely identifying the record.
        name : str, optional
            The name to use in uniquely identifying the record.
        tar : bytes, optional
            The bytes content of a tar file to save.  tar cannot be given
            with root_dir.
        root_dir : str, optional
            Specifies the root directory for finding the directory to archive.
            The directory to archive is at <root_dir>/<name>.  (Default is to
            set root_dir to the current working directory.)  tar cannot be given
            with root_dir.
        
        Raises
        ------
        ValueError
            If style and/or name content given with record or the record already
            has an archive.
        """
        raise AttributeError('add_tar not defined for Database style')

    def update_tar(self,
                   record: Optional[Record] = None,
                   style: Optional[str] = None,
                   name: Optional[str] = None,
                   tar: Optional[bytes] = None,
                   root_dir: Optional[Path] = None):
        """
        Replaces an existing tar archive for a record with a new one.
        
        Parameters
        ----------
        record : Record, optional
            The record to associate the tar archive with.  If not given, then 
            name and/or style necessary to uniquely identify the record are 
            needed.
        style : str, optional
            The style to use in uniquely identifying the record.
        name : str, optional
            The name to use in uniquely identifying the record.
        tar : bytes, optional
            The bytes content of a tar file to save.  tar cannot be given
            with root_dir.
        root_dir : str, optional
            Specifies the root directory for finding the directory to archive.
            The directory to archive is at <root_dir>/<name>.  (Default is to
            set root_dir to the current working directory.)  tar cannot be given
            with root_dir.
        
        Raises
        ------
        AttributeError
            If update_tar is not defined for database style.
        """
        raise AttributeError('update_tar not defined for Database style')

    def delete_tar(self, 
                   record: Optional[Record] = None,
                   style: Optional[str] = None,
                   name: Optional[str] = None):
        """
        Deletes a tar file from the database.
        
        Parameters
        ----------
        record : Record, optional
            The record associated with the tar archive to delete.  If not
            given, then name and/or style are necessary to uniquely identify
            the record.
        style : str, optional
            The style to use in uniquely identifying the record.
        name : str, optional
            The name to use in uniquely identifying the record.
        
        Raises
        ------
        AttributeError
            If delete_tar is not defined for database style.
        """
        raise AttributeError('delete_tar not defined for Database style')

    def copy_records(self,
                     dest,
                     record_style: Optional[str]  = None,
                     records: Optional[list] = None,
                     includetar: bool = True,
                     overwrite: bool = False):
        """
        Copies records from the current database to another database.
        
        Parameters
        ----------
        dest :  Database
            The destination database to copy records to.
        record_style : str, optional
            The record style to copy.  If record_style and records not
            given, then the available record styles will be listed and the
            user prompted to pick one.  Cannot be given with records.
        records : list, optional
            A list of Record objects from the current database to copy
            to dbase2.  Allows the user full control on which records to
            copy/update.  Cannot be given with record_style.
        includetar : bool, optional
            If True, the tar archives will be copied along with the records.
            If False, only the records will be copied. (Default is True).
        overwrite : bool, optional
            If False (default) only new records and tars will be copied.
            If True, all existing content will be updated.
        """
        if record_style is None and records is None:
            # Prompt for record_style
            record_style = self.select_record_style()

        if record_style is not None:
            if records is not None:
                raise ValueError('record_style and records cannot both be given')

            # Retrieve records from self
            records = self.get_records(style=record_style)

        elif records is None:
            # Set empty list if record_style is still None and no records given
            records = []

        print(len(records), 'records to try to copy')

        record_count = 0
        tar_count = 0
        # Copy records
        for record in tqdm(records, 'copying records', ascii=True):
            try:
                # Add new records
                dest.add_record(record=record)
                record_count += 1
            except:
                # Update existing records
                if overwrite:
                    dest.update_record(record=record)
                    record_count += 1

            # Copy archives
            if includetar:
                try:
                    # Get tar if it exists
                    tar = self.get_tar(record=record, raw=True) 
                except:

                    # Get folder if it exists
                    try:
                        root_dir = self.get_folder(record=record).parent
                    except:
                        pass
                    else:
                        try:
                            # Copy tar over
                            dest.add_tar(record=record, root_dir=root_dir)
                            tar_count += 1
                        except:
                            # Update existing tar
                            if overwrite:
                                dest.update_tar(record=record, root_dir=root_dir)
                                tar_count += 1

                else:
                    try:
                        # Copy tar over
                        dest.add_tar(record=record, tar=tar)
                        tar_count += 1
                    except:
                        # Update existing tar
                        if overwrite:
                            dest.update_tar(record=record, tar=tar)
                            tar_count += 1

        print(record_count, 'records added/updated')
        if includetar:
            print(tar_count, 'tars added/updated')

    def destroy_records(self,
                        record_style: Optional[str] = None,
                        records: Optional[list] = None,
                        prompt: bool = True):
        """
        Permanently deletes multiple records and their associated tars all at
        once.
        
        Parameters
        ----------
        record_style : str, optional
            The record style to delete.  If given, all records of that style
            will be deleted. If neither record_style nor records given, then
            the available record styles will be listed and the user prompted
            to pick one.
        records : list, optional
            A list of pre-selected records to delete.
        prompt : bool, optional
            If True (default) a prompt will ask if you are certain you want
            to destroy the matching records.  Only set to False if you are
            absolutely sure your code is deleting the correct records.
        """
        # Select record_style if needed
        if record_style is None and records is None:
            record_style = self.select_record_style()

        # Get records by record_style
        if record_style is not None:
            if records is not None:
                raise ValueError('record_style and records cannot both be given')

            # Retrieve records with errors from self
            records = self.get_records(style=record_style) 

        elif records is None:
            # Set empty list if record_style is still None and no records given
            records = []

        print(f'{len(records)} records found to be destroyed')

        if len(records) > 0:
            if prompt:
                test = screen_input('Delete records? (must type yes):')
            else:
                test = 'yes'
            
            if test == 'yes':
                count = 0
                record_styles = set()
                for record in tqdm(records, 'destroying records', ascii=True):
                    try:
                        self.delete_tar(record=record)
                    except:
                        pass
                    try:
                        self.delete_record(record=record)
                        count += 1
                        record_styles.add(record.style)
                    except:
                        pass
                if self.style == 'local':
                    for record_style in record_styles:
                        cache = self.cache(record_style, refresh=True)
                print(count, 'records successfully deleted')

    def select_record_style(self) -> str:
        """
        Console prompt for selecting a record_style
        """
        # Build list of calculation records
        styles = recordmanager.loaded_style_names

        # Ask for selection
        print('Select record_style:')
        for i, style in enumerate(styles):
            print(i+1, style)
        choice = screen_input(':')
        try:
            choice = int(choice)
        except:
            record_style = choice
        else:
            record_style = styles[choice-1]
        print()

        return record_style

    def record_querynames(self,
                          style: str) -> list:
        """Return the list of query parameter names for a given record style"""
        return load_record(style).querynames

    def record_querydoc(self,
                        style: str) -> str:
        """Return the description of query parameters for a given record style"""
        return load_record(style).querydoc



# coding: utf-8
__all__ = ['Database', 'databasemanager', 'load_database']

# Import base Database class
from .Database import Database

# Initialize a ModuleManager for the database styles
from ..tools import ModuleManager
databasemanager = ModuleManager('Database')

# Import load_database
from .load_database import load_database

# Add the modular Database styles
databasemanager.import_style('local', '.LocalDatabase', __name__)
databasemanager.import_style('mongo', '.MongoDatabase', __name__)
databasemanager.import_style('cdcs', '.CDCSDatabase', __name__)


from typing import Optional, Union, Any

from DataModelDict import DataModelDict as DM

from ..query import load_query
from . import Value
from .. import unitconvert as uc

class FloatValue(Value):
    
    def __init__(self,
                 name: str,
                 record,
                 defaultvalue: Optional[Any] = None,
                 valuerequired: bool = False,
                 allowedvalues: Optional[tuple] = None,
                 metadatakey: Union[str, bool, None] = None,
                 metadataparent: Optional[str] = None,
                 modelpath: Optional[str] = None,
                 description: Optional[str] = None,
                 unit: Optional[str] = None):
        """
        Initialize a FloatValue object.

        Parameters
        ----------
        name : str
            The name of the parameter.  This should correspond to the name of
            the associated class attribute.
        record : Record
            The Record object that the Parameter is used with.
        defaultvalue : any or None, optional
            The default value to use for the property.  The default value of
            None indicates that there is no default value.
        valuerequired: bool
            Indicates if a value must be given for the property.  If True, then
            checks will be performed that a value is assigned to the property.
        allowedvalues : tuple or None, optional
            A list/tuple of values that the parameter is restricted to have.
            Setting this to None (default) indicates any value is allowed.
        metadatakey: str, bool or None, optional
            The key name to use for the property when constructing the record
            metadata dict.  If set to None (default) then name will be used for
            metadatakey.  If set to False then the parameter will not be
            included in the metadata dict.
        metadataparent: str or None, optional
            If given, then this indicates that the metadatakey is actually an
            element of a dict in metadata with this name.  This allows for limited
            support for metadata having embedded dicts.
        modelpath: str, optional
            The period-delimited path after the record root element for
            where the parameter will be found in the built data model.  If set
            to None (default) then name will be used for modelpath.
        description: str or None, optional
            A short description for the value.  If not given, then the record name
            will be used.
        unit: str, optional
            The units to use when saving the value in a database.  A value of None
            (default) indicates that no unit conversions should be done.
        """
        self.__unit = unit
        
        super().__init__(name, record, defaultvalue=defaultvalue,
                         valuerequired=valuerequired, allowedvalues=allowedvalues,
                         metadatakey=metadatakey, metadataparent=metadataparent,
                         modelpath=modelpath, description=description)
        
    @property
    def unit(self) -> Optional[str]:
        """str or None: The units to use when saving the value to the model."""
        return self.__unit

    def set_value_mod(self, val):

        # Check if value is in #text
        val = self.set_value_mod_textfield(val)

        if val is None:
            return None
        elif isinstance(val, str) and self.unit is not None:
            return uc.set_literal(val)
        else:
            return float(val)
    
    @property
    def _default_queries(self) -> dict:
        """dict: Default query operations to associate with the Parameter style"""
        if self.metadatakey is False:
            return {}
        else:
            return {
                self.name: load_query('float_match',
                                    name=self.metadatakey,
                                    parent=self.metadataparent,
                                    path=f'{self.record.modelroot}.{self.modelpath}',
                                    description=f'Return only the records where {self.description} matches a given value')
            }
    
    def build_model_value(self):
        if self.value is None:
            return None
        
        if self.unit is None:
            return self.value
        else:
            return uc.model(self.value, self.unit)
        
    def load_model_value(self, val):
        if self.unit is None:
            return val
        else:
            return uc.value_unit(val)
        
    def metadata(self, meta):
        """
        Adds the parameter to the record's metadata dict.

        Parameters
        ----------
        meta : dict
            The metadata dict being built for the record.
        """
        if self.metadatakey is False or self.value is None:
            return
        
        # Convert value to unit 
        value = self.value
        if self.unit is not None:
            value = uc.get_in_units(value, self.unit)
        
        if self.metadataparent is None:
            meta[self.metadatakey] = value

        else:
            if self.metadataparent not in meta:
                meta[self.metadataparent] = {}
            meta[self.metadataparent][self.metadatakey] = value
        


@property
def Hz(self) -> float:
    """hertz"""
    return 1. / self.s

@property
def mHz(self) -> float:
    """megahertz"""
    return 1e-3 * self.Hz

@property
def kHz(self) -> float:
    """kilohertz"""
    return 1e3 * self.Hz

@property
def MHz(self) -> float:
    """megahertz"""
    return 1e6 * self.Hz
    
@property
def GHz(self) -> float:
    """gigahertz"""
    return 1e9 * self.Hz

@property
def THz(self) -> float:
    """terahertz"""
    return 1e12 * self.Hz

@property
def PHz(self) -> float:
    """petahertz"""
    return 1e15 * self.Hz

@property
def rtHz(self) -> float:
    """root Hertz"""
    return self.Hz**0.5 

@property
def rpm(self) -> float:
    """revolutions per minute"""
    return 1 / self.minute

############ Angular frequency below ################

@property
def Hz·2π(self) -> float:
    """hertz times 2 pi"""
    return self.Hz * 2 * self.pi

@property
def mHz·2π(self) -> float:
    """millihertz times 2 pi"""
    return self.mHz * 2 * self.pi

@property
def kHz·2π(self) -> float:
    """kilohertz times 2 pi"""
    return self.kHz * 2 * self.pi

@property
def MHz·2π(self) -> float:
    """megahertz times 2 pi"""
    return self.MHz * 2 * self.pi

@property
def GHz·2π(self) -> float:
    """gigahertz times 2 pi"""
    return self.GHz * 2 * self.pi

@property
def THz·2π(self) -> float:
    """terahertz times 2 pi"""
    return self.THz * 2 * self.pi

@property
def PHz·2π(self) -> float:
    """petahertz times 2 pi"""
    return self.PHz * 2 * self.pi

@property
def rpm·2π(self) -> float:
    """revolutions per minute times 2 pi"""
    return self.rpm * 2 * self.pi


@property 
def W(self) -> float:
    """watt"""
    return self.J / self.s

@property
def mW(self) -> float:
    """milliwatt"""
    return 1e-3 * self.W

@property
def uW(self) -> float:
    """microwatt"""
    return 1e-6 * self.W

@property
def nW(self) -> float:
    """nanowatt"""
    return 1e-9 * self.W

@property
def pW(self) -> float:
    """picowatt"""
    return 1e-12 * self.W

@property
def kW(self) -> float:
    """kilowatt"""
    return 1e3 * self.W

@property
def MW(self) -> float:
    """megawatt"""
    return 1e6 * self.W

@property
def GW(self) -> float:
    """gigawatt"""
    return 1e9 * self.W

@property
def TW(self) -> float:
    """terawatt"""
    return 1e12 * self.W

@property
def horsepower_imperial(self) -> float:
    """imperial horsepower"""
    return 33000 * self.foot * self.lbf / self.minute

@property
def horsepower_metric(self) -> float:
    """metric horsepower"""
    return (75 * self.kg) * (9.80665 * self.m / self.s**2) * (1 * self.m / self.s)


@property
def N(self) -> float:
    """newton"""
    return (self.kg * self.m) / self.s**2

@property
def mN(self) -> float:
    """millinewton"""
    return 1e-3 * self.N

@property
def uN(self) -> float:
    """micronewton"""
    return 1e-6 * self.N

@property
def nN(self) -> float:
    """nanonewton"""
    return 1e-9 * self.N

@property
def pN(self) -> float:
    """piconewton"""
    return 1e-12 * self.N

@property
def fN(self) -> float:
    """femtonewton"""
    return 1e-15 * self.N

@property
def kN(self) -> float:
    """kilonewton"""
    return 1e3 * self.N

@property
def MN(self) -> float:
    """meganewton"""
    return 1e6 * self.N

@property
def GN(self) -> float:
    """giganewton"""
    return 1e9 * self.N

@property
def dyn(self) -> float:
    """dyne"""
    return 1e-5 * self.N

@property
def lbf(self) -> float:
    """pound-force (international avoirdupois pound)"""
    return self.lbm * (9.80665 * self.m / self.s**2)